@@ -4,7 +4,9 @@ import re
 from pathlib import Path
 
 import jsonschema
+import numpy as np
 import pandas as pd
+from dask import compute, delayed
 
 RECIPE_SCHEMA = {
     "type": "object",
@@ -41,15 +43,52 @@ RECIPE_SCHEMA = {
 }
 
 
+@delayed
+def min_max_scale(s):
+    min = np.nanmin(s)
+    max = np.nanmax(s)
+    if max == min:
+        exit(
+            " ❌ Scaling failed: There is at least one constant column.\n   Consider dropping constant columns, disabling or skip scaling."
+        )
+    return (s - min) / (max - min)
+
+
+@delayed
+def min_max_scale_limits(s, min_limit, max_limit):
+    return (s - min_limit) / (max_limit - min_limit)
+
+
 class ProgressMessage:
-    def __init__(self, message):
+    def __init__(self, message, parent=None):
         self.message = message
+        self.children = []
+        self.level = 0 if parent is None else parent.level + 1
+        if parent is not None:
+            if len(parent.children) == 0:
+                print("", end="\n")
+            parent.children.append(self)
 
     def __enter__(self):
-        print(f"⬜ {self.message}", end="\r")
+        print("\033[?25l", end="")
+        print("  " * self.level, end="")
+        print(f" ⬜  {self.message}", end="\r")
+        print("\b" * 10, end="")
+        return self
 
     def __exit__(self, exc_type, exc_value, traceback):
-        print(f"✅ {self.message}", end="\r\n")
+        if len(self.children) > 0:
+            self._clear_line(len(self.children) + 1)
+        print("  " * self.level, end="")
+        print(f" ✅  {self.message}", end="\r\n")
+        for child in self.children:
+            print("  " * (child.level), end="")
+            print(f" ✅  {child.message}", end="\r\n")
+        print("\033[?25h", end="")
+
+    def _clear_line(self, n=1):
+        for _ in range(n):
+            print("\033[1A", end="\x1b[2K")
 
 
 class DataReleasePrep:
@@ -57,7 +96,6 @@ class DataReleasePrep:
         self,
         recipe_file,
         input_file,
-        output_file,
         limits_file,
         dry_run,
         verbose,
@@ -66,7 +104,6 @@ class DataReleasePrep:
     ):
         self.recipe_file = recipe_file
         self.input_file = input_file
-        self.output_file = output_file
         self.limits_file = limits_file
         self.dry_run = dry_run
         self.verbose = verbose
@@ -74,6 +111,7 @@ class DataReleasePrep:
         self.limits = None
         self.report = []
 
+        self.output_directory = str(Path(self.recipe_file).parent.absolute()) + "/"
         self.input_file_stem = Path(self.input_file).stem
         self.input_file_suffix = Path(self.input_file).suffix
 
@@ -97,14 +135,13 @@ class DataReleasePrep:
         self.recipe = json.load(open(self.recipe_file))
         jsonschema.validate(self.recipe, RECIPE_SCHEMA)
 
-        # Define the output file if not given
-        if self.output_file is None:
-            self.output_file = (
-                self.input_file_stem
-                + "_release_"
-                + self.recipe["version"]
-                + self.input_file_suffix
-            )
+        self.output_file = (
+            self.output_directory
+            + self.input_file_stem
+            + "_release_"
+            + self.recipe["version"]
+            + self.input_file_suffix
+        )
         self._report_log("recipe_version", "", self.recipe["version"])
 
     def _read_limits(self):
@@ -169,38 +206,115 @@ class DataReleasePrep:
                                 )
 
     def _scale_columns(self):
-        with ProgressMessage("Scaling columns..."):
-            for col in self.data.select_dtypes(include="number").columns.tolist():
-                if col in self.recipe["actions"]["obfuscate"]:
-                    continue
+        with ProgressMessage("Scaling columns...") as level1:
+            min_max_scale_limit_cols = []
+            min_max_scale_cols = []
+            min_max_scale_limit_futures = []
+            min_max_scale_futures = []
+            with ProgressMessage("Preparing compute processes...", parent=level1):
+                for col in self.data.select_dtypes(include="number").columns.tolist():
+                    if col in self.recipe["actions"].get("obfuscate", []):
+                        continue
 
-                skip_scaling = False
-                no_scaling = self.recipe["actions"].get("skip-scaling", [])
-                for pat in no_scaling:
-                    if re.fullmatch(pat, col):
-                        skip_scaling = True
-                        break
+                    skip_scaling = False
+                    no_scaling = self.recipe["actions"].get("skip-scaling", [])
+                    for pat in no_scaling:
+                        if re.fullmatch(pat, col):
+                            skip_scaling = True
+                            break
 
-                if not skip_scaling:
-                    if self.limits is not None and col in self.limits:
-                        min, max = self.limits[col]["min"], self.limits[col]["max"]
-                        if pd.isna(min):
-                            min = self.data[col].min()
-                        if pd.isna(max):
-                            max = self.data[col].max()
-                        self._report_log("SCALE_CUSTOM", col, f"[{min},{max}]")
-                        if not self.dry_run:
-                            self.data[col] = (self.data[col] - min) / (max - min)
-                    else:
-                        self._report_log(
-                            "SCALE_DEFAULT",
-                            col,
-                            f"[{self.data[col].min()},{self.data[col].max()}]",
-                        )
-                        if not self.dry_run:
-                            self.data[col] = (self.data[col] - self.data[col].min()) / (
-                                self.data[col].max() - self.data[col].min()
+                    if not skip_scaling:
+                        col_min = self.data[col].min()
+                        col_max = self.data[col].max()
+                        if self.limits is not None and col in self.limits:
+                            min, max = self.limits[col]["min"], self.limits[col]["max"]
+                            if pd.isna(min):
+                                min = col_min
+                            if pd.isna(max):
+                                max = col_max
+                            self._report_log("SCALE_CUSTOM", col, f"[{min},{max}]")
+                            if not self.dry_run:
+                                min_max_scale_limit_cols.append(col)
+                                if self.data[col].dtype.name == "int64":
+                                    min_max_scale_limit_futures.append(
+                                        min_max_scale_limits(
+                                            self.data[col].to_numpy(
+                                                dtype=pd.Int64Dtype, na_value=np.nan
+                                            )
+                                        )
+                                    )
+                                else:
+                                    min_max_scale_limit_futures.append(
+                                        min_max_scale_limits(
+                                            self.data[col].to_numpy(na_value=np.nan),
+                                            min,
+                                            max,
+                                        )
+                                    )
+                        else:
+                            self._report_log(
+                                "SCALE_DEFAULT",
+                                col,
+                                f"[{col_min},{col_max}]",
                             )
+                            if not self.dry_run:
+                                min_max_scale_cols.append(col)
+                                if self.data[col].dtype.name == "int64":
+                                    min_max_scale_futures.append(
+                                        min_max_scale(
+                                            self.data[col].to_numpy(
+                                                dtype=pd.Int64Dtype, na_value=np.nan
+                                            )
+                                        )
+                                    )
+                                else:
+                                    min_max_scale_futures.append(
+                                        min_max_scale(
+                                            self.data[col].to_numpy(na_value=np.nan)
+                                        )
+                                    )
+
+            if not self.dry_run:
+                if len(min_max_scale_limit_futures) > 0:
+                    with ProgressMessage(
+                        "Running limit scaling processes...",
+                        parent=level1,
+                    ):
+                        computed_columns = compute(
+                            *min_max_scale_limit_futures, scheduler="processes"
+                        )
+                        self.data.drop(min_max_scale_limit_cols, axis=1, inplace=True)
+                        computed_columns = [
+                            pd.Series(computed_column)
+                            for computed_column in computed_columns
+                        ]
+                        computed_columns = pd.concat(computed_columns, axis=1)
+                        computed_columns.columns = min_max_scale_limit_cols
+                        self.data = self.data.merge(
+                            computed_columns,
+                            left_index=True,
+                            right_index=True,
+                        )
+
+                if len(min_max_scale_futures) > 0:
+                    with ProgressMessage(
+                        "Running default min/max scaling processes...", parent=level1
+                    ):
+                        computed_columns = compute(
+                            *min_max_scale_futures, scheduler="processes"
+                        )
+                        self.data.drop(min_max_scale_cols, axis=1, inplace=True)
+                        computed_columns = [
+                            pd.Series(computed_column)
+                            for computed_column in computed_columns
+                        ]
+                        computed_columns = pd.concat(computed_columns, axis=1)
+                        computed_columns.columns = min_max_scale_cols
+                        self.data = self.data.merge(
+                            computed_columns,
+                            left_index=True,
+                            right_index=True,
+                        )
 
     def _rename_columns(self):
         if "rename" in self.recipe["actions"]:
@@ -245,20 +359,25 @@ class DataReleasePrep:
         self._rename_columns()
         if not self.dry_run:
             if self.input_file_suffix == ".csv":
-                self.data.to_csv(
-                    self.output_file,
-                    index=False,
-                )
+                with ProgressMessage("Generating data release CSV file..."):
+                    self.data.to_csv(
+                        self.output_file,
+                        index=False,
+                    )
             elif self.input_file_suffix == ".parquet":
-                self.data.to_parquet(
-                    self.output_file,
-                    engine="pyarrow",
-                    index=False,
-                )
+                with ProgressMessage("Generating data release Parquet file..."):
+                    self.data.to_parquet(
+                        self.output_file,
+                        engine="pyarrow",
+                        index=False,
+                    )
 
     def generate_report(self):
         with ProgressMessage("Generating report..."):
             report_df = pd.DataFrame(
                 self.report, columns=["action", "column", "details"]
             )
-            report_df.to_csv(Path(self.output_file).stem + "_report.csv", index=True)
+            report_df.to_csv(
+                self.output_directory + Path(self.output_file).stem + "_report.csv",
+                index=True,
+            )
